from typing import Callable, List

import torch


class BatchedObjective:
    def __init__(self, objectives: List[Callable]) -> None:
        """Wrapper to handle generations of multiple feature visualizations batch-wise.

        Args:
            objectives (List[Callable]): list of all custom_func i.e objective functions. It's length should be same as that of the batch size.
        """
        self.objectives = objectives

    def __call__(self, x: list):

        for y in x:
            assert torch.is_tensor(
                y
            ), f"Expected every item in x to be a torch.tensor, but got: {type(y)}"

        batch_size = x[0].shape[0]

        assert batch_size == len(
            self.objectives
        ), f"Batch size of x and len(self.objectives) do not match: {x[0].shape[0]} and {len(self.objectives)}"

        loss_sum = torch.tensor(0.0).to(x[0].device)

        for batch_idx in range(batch_size):
            loss_sum += self.objectives[batch_idx](
                x[0][batch_idx].unsqueeze(0)
            )  ## is unsqueeze this really necessary? idk

        return loss_sum  ## then call loss_sum.backward()



import torch
import torchvision
import torch.nn as nn
from tqdm import tqdm
from copy import deepcopy
import torchvision.transforms as transforms

from .transforms import random_resize, pair_random_resize, pair_random_affine

from .constants import Constants
from .losses import CaricatureLoss
from .image_transforms import InverseTransform
from .auto_image_param import AutoImageParam
from .dreamer_utils import Hook, default_func_mean
from .masked_image_param import MaskedImageParam
from .batched_image_param import BatchedImageParam


class Dreamer:
    """wrapper over a pytorch model for feature visualization

    Args:
        model (torch.nn.Module): pytorch model
        quiet (bool, optional): enable or disable progress bar. Defaults to True.
        device (str, optional): 'cpu' or 'cuda'. Defaults to 'cuda'.

    Example:

    ```python
    import torchvision.models as models
    from torch_dreams import Dreamer

    model = models.inception_v3(pretrained=True)
    dreamy_boi = Dreamer(model, device = 'cuda')
    ```
    """

    def __init__(self, model, quiet=False, device="cuda"):

        self.model = model
        self.model.eval()
        self.device = device
        self.model.to(self.device)
        self.default_func = default_func_mean
        self.transforms = None
        self.quiet = quiet
        self.__custom_normalization_transform__ = None

    def get_default_transforms(
        self, rotate, scale_max, scale_min, translate_x, translate_y
    ):
        self.transforms = transforms.Compose(
            [
                transforms.RandomAffine(
                    degrees=rotate, translate=(translate_x, translate_y)
                ),
                random_resize(max_size_factor=scale_max, min_size_factor=scale_min),
            ]
        )

        if self.__custom_normalization_transform__ is not None:
            self.transforms = transforms.Compose(
                [
                    transforms.RandomAffine(
                        degrees=rotate, translate=(translate_x, translate_y)
                    ),
                    random_resize(max_size_factor=scale_max, min_size_factor=scale_min),
                    self.__custom_normalization_transform__,
                ]
            )

    def set_custom_normalization(self, normalization_transform):
        """Overrides the usual imagenet normalization to be compatible with any other normalization defined by the user.

        Args:
            normalization_transform (torch.nn.Module): transforms that would normalize the image as per your model.

        Example:

        ```python
        t = torchvision.transforms.Normalize(
                mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]
            )

        dreamer.set_custom_normalization(normalization_transform = t)
        ```
        """

        assert isinstance(normalization_transform, nn.Module) == True, (
            "normalization_transform should be an instance of torch.nn.module, not: "
            + str(type(normalization_transform))
        )

        self.__custom_normalization_transform__ = InverseTransform(
            old_mean=Constants.imagenet_mean,
            old_std=Constants.imagenet_std,
            new_transforms=normalization_transform,
        )

    def set_custom_transforms(self, transforms):
        """Sets a custom set of transforms as the robustness transforms to be used for training the image parameter. Does not override custom image normalization if configured.

        Args:
            transforms (torch.nn.Module): Bunch of transforms to be used for training the image parameter. Generally `torchvision.transforms.Compose([your_transforms])` is preferred.

        Example:

        ```python
        my_transforms = transforms.Compose([
            transforms.RandomAffine(degrees = 10, translate = (0.5,0.5)),
            transforms.RandomHorizontalFlip(p = 0.3)
        ])

        dreamy_boi.set_custom_transforms(transforms = my_transforms)
        ```
        """
        if self.__custom_normalization_transform__ == None:
            self.transforms = transforms
        else:
            self.transforms = torchvision.transforms.Compose(
                [transforms, self.__custom_normalization_transform__]
            )

    def render(
        self,
        layers,
        image_parameter=None,
        width=256,
        height=256,
        iters=120,
        lr=9e-3,
        rotate_degrees=15,
        scale_max=1.2,
        scale_min=0.5,
        translate_x=0.0,
        translate_y=0.0,
        custom_func=None,
        weight_decay=0.0,
        grad_clip=1.0,
    ):
        """core function to visualize elements form within the pytorch model

        WARNING: width and height would be ignored if image_parameter is not None

        Args:
            layers (iterable): [model.layer1, model.layer2...]
            image_parameter: instance of torch_dreams.auto.auto_image_param
            width (int): width of image to be optimized
            height (int): height of image to be optimized
            iters (int): number of iterations, higher -> stronger visualization
            lr (float): learning rate
            rotate_degrees (int): max rotation in default transforms
            scale_max (float, optional): Max image size factor. Defaults to 1.1.
            scale_min (float, optional): Minimum image size factor. Defaults to 0.5.
            translate_x (float, optional): Maximum translation factor in x direction
            translate_y (float, optional): Maximum translation factor in y direction
            custom_func (function, optional): See docs for a better explanation. Defaults to None.
            weight_decay (float, optional): Weight decay for default optimizer. Helps prevent high frequency noise. Defaults to 0.
            grad_clip (float, optional): Maximum value of norm of gradient. Defaults to 1.

        Returns:
            image_parameter instance: To show image, use: plt.imshow(image_parameter)
        """
        if image_parameter is None:
            image_parameter = AutoImageParam(
                height=height, width=width, device=self.device, standard_deviation=0.01
            )
        else:
            image_parameter = deepcopy(image_parameter)

        if image_parameter.optimizer is None:
            image_parameter.get_optimizer(lr=lr, weight_decay=weight_decay)

        if self.transforms is None:
            self.get_default_transforms(
                rotate=rotate_degrees,
                scale_max=scale_max,
                scale_min=scale_min,
                translate_x=translate_x,
                translate_y=translate_y,
            )

        hooks = []
        for layer in layers:
            hook = Hook(layer)
            hooks.append(hook)

        if isinstance(image_parameter, MaskedImageParam):
            self.random_resize_pair = pair_random_resize(
                max_size_factor=scale_max, min_size_factor=scale_min
            )
            self.random_affine_pair = pair_random_affine(
                degrees=rotate_degrees, translate_x=translate_x, translate_y=translate_y
            )

        for i in tqdm(range(iters), disable=self.quiet):

            image_parameter.optimizer.zero_grad()

            img = image_parameter.forward(device=self.device)

            if isinstance(image_parameter, MaskedImageParam):
                (
                    img_transformed,
                    mask_transformed,
                    original_image_transformed,
                ) = self.random_resize_pair(
                    tensors=[
                        img,
                        image_parameter.mask.to(self.device),
                        image_parameter.original_nchw_image_tensor,
                    ]
                )
                (
                    img_transformed,
                    mask_transformed,
                    original_image_transformed,
                ) = self.random_affine_pair(
                    [img_transformed, mask_transformed, original_image_transformed]
                )

                img = img_transformed * mask_transformed.to(
                    self.device
                ) + original_image_transformed.float() * (
                    1 - mask_transformed.to(self.device)
                )

            else:
                img = self.transforms(img)

            model_out = self.model(img)

            layer_outputs = []

            for hook in hooks:
                ## if it's a BatchedImageParam, then include all batch items from hook output
                if isinstance(image_parameter, BatchedImageParam):
                    out = hook.output
                else:
                    ## else select only the first and only batch item
                    out = hook.output[0]

                layer_outputs.append(out)

            if custom_func is not None:
                loss = custom_func(layer_outputs)
            else:
                loss = self.default_func(layer_outputs)
            loss.backward()
            image_parameter.clip_grads(grad_clip=grad_clip)
            image_parameter.optimizer.step()

        for hook in hooks:
            hook.close()

        return image_parameter

    def get_snapshot(self, layers, input_tensor):
        """Registers the outputs of a set of layers within a model

        Args:
            layers (list): [model.layer1, model.layer2,...]
            input_tensor (torch.Tensor): NCHW tensor to be fed into the model

        Returns:
            [list]: list of layer outputs
        """

        with torch.no_grad():
            input_tensor = input_tensor.to(self.device)

            hooks = []
            for layer in layers:
                hook = Hook(layer)
                hooks.append(hook)

            model_out = self.model(input_tensor.float())

            layer_outputs = []

            for hook in hooks:
                out = hook.output[0]
                layer_outputs.append(out)

        return layer_outputs

    def caricature(
        self,
        input_tensor,
        layers,
        power=1.0,
        image_parameter=None,
        iters=120,
        lr=9e-3,
        rotate_degrees=15,
        scale_max=1.2,
        scale_min=0.5,
        translate_x=0.1,
        translate_y=0.1,
        weight_decay=1e-3,
        grad_clip=1.0,
        static=False,
    ):
        """Generates an "exaggerated" reconstruction of the input image by
        optimizing random noise to replicate and "exaggerate" the activations of
        certain layers after feeding the input image

        Args:
            input_tensor (torch.tensor): input image tensor, should be clipped between 0,1 and Normalised w.r.t image mean and std
            layers (list): list of layers whose activations are to be registered and later replicated on random noise
            power (float, optional): Determined the "exaggeration" factor. 1 means "no exaggeration". Defaults to 1..
            image_parameter ([type], optional): [description]. Defaults to None.
            iters (int, optional): Number of optimization steps. Defaults to 120.
            lr (float, optional): learning rate. Defaults to 9e-3.
            rotate_degrees (int, optional): Maximum amount of random rotation in degrees. Defaults to 15.
            scale_max (float, optional): Maximum scale factor for random scaling. Defaults to 1.2.
            scale_min (float, optional): Minimum scale factor for random scaling. Defaults to 0.5.
            translate_x (float, optional): Maximum amount of horizontal jitter. Defaults to 0.1.
            translate_y (float, optional): Maximum amount of vertical jitter. Defaults to 0.1.
            weight_decay (float, optional): Helps reduce high frequency noise, higher means lower noise. Defaults to 1e-3.
            grad_clip (float, optional): Maximum value of grad norm. Smaller -> more "careful" steps. Defaults to 1..

        Example:
        ```python
        param = dreamy_boi.caricature(
            input_tensor = image_tensor,
            layers = [model.some_layer],
            power= 1.0
        )
        param.save('caricature.jpg')
        ```

        Returns:
            instance of torch_dreams.auto_image_param.BaseImageParam
        """

        if image_parameter is None:
            height, width = input_tensor.shape[-2], input_tensor.shape[-1]
            image_parameter = AutoImageParam(
                height=height, width=width, device=self.device, standard_deviation=0.01
            )
        else:
            image_parameter = deepcopy(image_parameter)

        if image_parameter.optimizer is None:
            image_parameter.get_optimizer(lr=lr, weight_decay=weight_decay)

        hooks = []
        for layer in layers:
            hook = Hook(layer)
            hooks.append(hook)

        caricature_loss = CaricatureLoss(power=power)

        if static == False:
            """
            if static is false, we have to transforms the input tensor and the image parameter identically before each iteration
            """
            self.random_resize_pair = pair_random_resize(
                max_size_factor=scale_max, min_size_factor=scale_min
            )
            self.random_affine_pair = pair_random_affine(
                degrees=rotate_degrees, translate_x=translate_x, translate_y=translate_y
            )
        else:
            """
            if static is true, then we collect the layer outputs first, and use the same ideal layer outputs as the target at each iteration,
            this would be useful particularly for adversarial examples where the outcome depends on the position of each pixel
            """
            self.get_default_transforms(
                rotate=rotate_degrees,
                scale_max=scale_max,
                scale_min=scale_min,
                translate_x=translate_x,
                translate_y=translate_y,
            )
            model_out = self.model(input_tensor.to(self.device))

            ideal_layer_outputs = []

            for hook in hooks:
                out = hook.output[0]
                ideal_layer_outputs.append(out)

        for i in tqdm(range(iters), disable=self.quiet):

            image_parameter.optimizer.zero_grad()

            img = image_parameter.forward(device=self.device)

            if static == False:
                img_transformed, input_tensor_transformed = self.random_resize_pair(
                    tensors=[img, input_tensor]
                )
                img_transformed, input_tensor_transformed = self.random_affine_pair(
                    tensors=[img_transformed, input_tensor_transformed]
                )
            else:
                img_transformed = self.transforms(img)

            model_out = self.model(img_transformed)

            layer_outputs = []

            for hook in hooks:
                out = hook.output[0]
                layer_outputs.append(out)

            if static == False:
                model_out = self.model(input_tensor_transformed.to(self.device))

                ideal_layer_outputs = []

                for hook in hooks:
                    out = hook.output[0]
                    ideal_layer_outputs.append(out)

            loss = caricature_loss.forward(
                layer_outputs=layer_outputs, ideal_layer_outputs=ideal_layer_outputs
            )

            loss.backward()
            image_parameter.clip_grads(grad_clip=grad_clip)
            image_parameter.optimizer.step()

        for hook in hooks:
            hook.close()

        return image_parameter



import torchvision.transforms as transforms
import torch
import torch.nn.functional as F
import torch.nn as nn


import warnings

warnings.filterwarnings("ignore")

transform_to_tensor = transforms.Compose([transforms.ToTensor()])

transform_and_rotate = transforms.Compose(
    [
        transforms.ToPILImage(),
        transforms.RandomRotation(
            degrees=5, expand=False, center=None, fill=None
        ),
        transforms.ToTensor(),
    ]
)


class InverseTransform(nn.Module):
    def __init__(self, old_mean, old_std, new_transforms):
        super().__init__()
        self.inverse_transform = transforms.Compose(
            [
                transforms.Normalize(
                    mean=[0.0, 0.0, 0.0],
                    std=[1 / old_std[0], 1 / old_std[1], 1 / old_std[2]],
                ),
                transforms.Normalize(
                    mean=[-old_mean[0], -old_mean[1], -old_mean[2]], std=[1.0, 1.0, 1.0]
                ),
            ]
        )

        self.new_transform = new_transforms

    def forward(self, x):
        x = self.inverse_transform(x)
        x = self.new_transform(x)
        return x

    def __repr__(self):
        return (
            "InverseTransform(\n        "
            + str(self.inverse_transform)
            + "\n        "
            + str(self.new_transform)
            + ")"
        )


def resize_4d_tensor_by_factor(x, height_factor, width_factor):
    res = F.interpolate(x, scale_factor=(height_factor, width_factor), mode="bilinear")
    return res


def resize_4d_tensor_by_size(x, height, width):
    res = F.interpolate(x, size=(height, width), mode="bilinear")
    return res



import torch.nn as nn
import torch
import numpy as np

from .utils import init_image_param
import torchvision.transforms as transforms

from .utils import fft_to_rgb, lucid_colorspace_to_rgb, normalize


class BaseImageParam(nn.Module):
    def __init__(self):
        super().__init__()
        self.height = None
        self.width = None
        self.param = None
        self.optimizer = None

    def forward(self):
        """This is what the model gets, should be processed and normalized with the right values

        The model gets:  self.normalize(self.postprocess(self.param))

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError

    def postprocess(self):
        """Moves the image from the frequency domain to Spatial (Visible to the eyes)

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError

    def normalize(self):
        """Normalizing wrapper, you can either use torchvision.transforms.Normalize() or something else

        Raises:
            NotImplementedError: Implemented below, you're in the base class.
        """
        raise NotImplementedError

    def fetch_optimizer(self, params_list, optimizer=None, lr=1e-3, weight_decay=0.0):
        if optimizer is not None:
            optimizer = optimizer(params_list, lr=lr, weight_decay=weight_decay)
        else:
            optimizer = torch.optim.AdamW(params_list, lr=lr, weight_decay=weight_decay)
        return optimizer

    def get_optimizer(self, lr, weight_decay):

        self.optimizer = self.fetch_optimizer(
            params_list=[self.param], lr=lr, weight_decay=weight_decay
        )

    def clip_grads(self, grad_clip=1.0):
        torch.nn.utils.clip_grad_norm_(self.param, grad_clip)

    def to_hwc_tensor(self, device="cpu"):
        rgb = self.postprocess(device=device)[0].permute(1, 2, 0).detach()
        return rgb

    def to_chw_tensor(self, device="cpu"):
        t = self.postprocess(device=device)[0].detach()
        return t

    def __array__(self):
        """Generally used for plt.imshow(), converts the image parameter to a numpy array

        Returns:
            numpy.ndarray
        """
        return self.to_hwc_tensor().numpy()

    def save(self, filename):
        """Save an image_param as an image. Uses PIL to save the image

        usage:

            image_param.save(filename = 'my_image.jpg')

        Args:
            filename (str): image.jpg
        """
        ten = self.to_chw_tensor()
        im = transforms.ToPILImage()(ten)
        im.save(filename)


class AutoImageParam(BaseImageParam):
    """Trainable image parameter which can be used to activate
       different parts of a neural net

    Args:
        height (int): Height of image
        width (int): Width of image
        device (str): 'cpu' or 'cuda'
        standard_deviation (float): Standard deviation of the image initiated in the frequency domain. ). 0.01 works well
    """

    def __init__(self, height, width, device, standard_deviation):

        super().__init__()
        self.height = height
        self.width = width

        """
        odd width is resized to even with one extra column
        """
        if self.width % 2 == 1:
            self.param = init_image_param(
                height=self.height,
                width=width + 1,
                sd=standard_deviation,
                device=device,
            )
        else:
            self.param = init_image_param(
                height=self.height,
                width=self.width,
                sd=standard_deviation,
                device=device,
            )

        self.param.requires_grad_()
        self.optimizer = None

    def postprocess(self, device):
        img = fft_to_rgb(
            height=self.height,
            width=self.width,
            image_parameter=self.param,
            device=device,
        )
        img = lucid_colorspace_to_rgb(t=img, device=device)
        img = torch.sigmoid(img)
        return img

    def normalize(self, x, device):
        return normalize(x=x, device=device)

    def forward(self, device):
        return self.normalize(self.postprocess(device=device), device=device)



import numpy as np
import torch


def color_correlation_normalized():
    color_correlation_svd_sqrt = np.asarray(
        [[0.26, 0.09, 0.02], [0.27, 0.00, -0.05], [0.27, -0.09, 0.03]]
    ).astype(np.float32)
    max_norm_svd_sqrt = np.max(np.linalg.norm(color_correlation_svd_sqrt, axis=0))
    color_correlation_normalized = torch.tensor(
        color_correlation_svd_sqrt / max_norm_svd_sqrt
    )
    return color_correlation_normalized


def imagenet_mean_std():
    return (torch.tensor([0.485, 0.456, 0.406]), torch.tensor([0.229, 0.224, 0.225]))


class Constants:
    color_correlation_matrix = color_correlation_normalized()
    imagenet_mean, imagenet_std = imagenet_mean_std()



import numpy as np
import torch
from torch import tensor
from torchvision import transforms

from .image_transforms import resize_4d_tensor_by_size
from .error_handlers import PytorchVersionError
from .constants import Constants


def init_image_param(height, width, sd=0.01, device="cuda"):
    """Initializes an image parameter in the frequency domain

    Args:
        height (int): height of image
        width (int): width of image
        sd (float, optional): Standard deviation of pixel values. Defaults to 0.01.
        device (str): 'cpu' or 'cuda'

    Returns:
        torch.tensor: image param to backpropagate on
    """
    img_buf = np.random.normal(size=(1, 3, height, width), scale=sd).astype(np.float32)
    spectrum_t = tensor(img_buf).float().to(device)
    return spectrum_t


def get_fft_scale(h, w, decay_power=0.75, device="cuda"):
    d = 0.5**0.5  # set center frequency scale to 1
    fy = np.fft.fftfreq(h, d=d)[:, None]

    if w % 2 == 1:
        fx = np.fft.rfftfreq(w, d=d)[: (w + 1) // 2]
    else:

        fx = np.fft.rfftfreq(w, d=d)[: w // 2]

    freqs = (fx * fx + fy * fy) ** decay_power
    scale = 1.0 / np.maximum(freqs, 1.0 / (max(w, h) * d))
    scale = tensor(scale).float().to(device)

    return scale


def fft_to_rgb(height, width, image_parameter, device="cuda"):
    """convert image param to NCHW

    WARNING: torch v1.7.0 works differently from torch v1.8.0 on fft.
    torch-dreams supports ONLY 1.8.x

    Latest docs: https://pytorch.org/docs/stable/fft.html

    Also refer:
        https://github.com/pytorch/pytorch/issues/49637

    Args:
        height (int): height of image
        width (int): width of image
        image_parameter (auto_image_param): auto_image_param.param

    Returns:
        torch.tensor: NCHW tensor

    """
    scale = get_fft_scale(height, width, device=device).to(image_parameter.device)
    # print(scale.shape, image_parameter.shape)
    if width % 2 == 1:
        image_parameter = image_parameter.reshape(1, 3, height, (width + 1) // 2, 2)
    else:
        image_parameter = image_parameter.reshape(1, 3, height, width // 2, 2)

    image_parameter = torch.complex(image_parameter[..., 0], image_parameter[..., 1])
    t = scale * image_parameter

    version = torch.__version__.split(".")[:2]
    main_version = int(version[0])
    sub_version = int(version[1])

    if (
        main_version >= 1 and sub_version >= 8
    ):  ## if torch.__version__ is greater than 1.8
        t = torch.fft.irfft2(t, s=(height, width), norm="ortho")
    else:
        raise PytorchVersionError(version=torch.__version__)

    return t


def lucid_colorspace_to_rgb(t, device="cuda"):

    t_flat = t.permute(0, 2, 3, 1)
    t_flat = torch.matmul(
        t_flat.to(device), Constants.color_correlation_matrix.T.to(device)
    )
    t = t_flat.permute(0, 3, 1, 2)
    return t


def normalize(x, device="cuda"):
    return (
        x - Constants.imagenet_mean[..., None, None].to(device)
    ) / Constants.imagenet_std[..., None, None].to(device)


def get_fft_scale_custom_img(h, w, decay_power=0.75, device="cuda"):
    d = 0.5**0.5  # set center frequency scale to 1
    fy = np.fft.fftfreq(h, d=d)[:, None]

    fx = np.fft.rfftfreq(w, d=d)[: (w // 2) + 1]
    freqs = (fx * fx + fy * fy) ** decay_power
    scale = 1.0 / np.maximum(freqs, 1.0 / (max(w, h) * d))
    scale = torch.tensor(scale).float().to(device)

    return scale


def denormalize(x):

    return x.float() * Constants.imagenet_std[..., None, None].to(
        x.device
    ) + Constants.imagenet_mean[..., None, None].to(x.device)


def rgb_to_lucid_colorspace(t, device="cuda"):
    t_flat = t.permute(0, 2, 3, 1)
    inverse = torch.inverse(Constants.color_correlation_matrix.T.to(device))
    t_flat = torch.matmul(t_flat.to(device), inverse)
    t = t_flat.permute(0, 3, 1, 2)
    return t


def chw_rgb_to_fft_param(x, device):
    im_tensor = torch.tensor(x).unsqueeze(0).float()

    x = rgb_to_lucid_colorspace(denormalize(im_tensor), device=device)

    x = torch.fft.rfft2(x, s=(x.shape[-2], x.shape[-1]), norm="ortho")
    return x


def fft_to_rgb_custom_img(height, width, image_parameter, device="cuda"):

    scale = get_fft_scale_custom_img(height, width, device=device).to(
        image_parameter.device
    )
    t = scale * image_parameter

    version = torch.__version__.split(".")[:2]
    main_version = int(version[0])
    sub_version = int(version[1])

    if (
        main_version >= 1 and sub_version >= 8
    ):  ## if torch.__version__ is greater than 1.8
        t = torch.fft.irfft2(t, s=(height, width), norm="ortho")
    else:
        raise PytorchVersionError(version=torch.__version__)

    return t



class PytorchVersionError(Exception):
    """Raised when the user is not on pytorch 1.8.x or higher

    Args:
        version (str): torch.__version__
    """

    def __init__(self, version):
        self.version = version
        self.message = (
            "Expected pytorch to have version 1.8.x or higher but got: "
            + self.version
            + "\n Please consider updating pytorch from: https://pytorch.org/get-started/locally"
        )

    def __str__(self):
        return self.message



import torch
import torch.nn as nn

import copy


class NoiseGradModule(nn.Module):
    """
    A tiny wrapper to apply noisegrad in nn.Module instances.

    Original paper: https://arxiv.org/pdf/2106.10185.pdf

    Original Implementation: https://github.com/understandable-machine-intelligence-lab/NoiseGrad/blob/master/src/noisegrad.py

    Args:
        module (nn.Module): a layer/module within the model
        mean (float): mean of the distribution from which we'd sample the noise values
        std (float): standard deviation of the distribution from which we'd sample the noise values

    Usage:

    ```python
    import torchvision.models as models
    from torch_dreams.noisegrad import NoiseGradModule

    model = models.inception_v3(pretrained=True)
    model.Mixed_5b = NoiseGradModule(module = model.Mixed_5b, mean = 1, std = 0.2)
    ```
    """

    def __init__(self, module: nn.Module, mean: float = 1, std: float = 0.2):
        super().__init__()

        self.mean = mean
        self.std = std

        # Creates a normal (also called Gaussian) distribution.
        self.distribution = torch.distributions.normal.Normal(
            loc=self.mean, scale=self.std
        )

        self.module = module

    @torch.no_grad()
    def sample(self):
        noisy_module = copy.deepcopy(self.module)

        for layer in noisy_module.parameters():
            noise = self.distribution.sample(layer.size()).to(layer.device)
            layer.mul_(noise)

        return noisy_module

    def forward(self, x):
        sampled_noisy_layer = self.sample()
        return sampled_noisy_layer.forward(x)



import torch
from typing import List
from .auto_image_param import BaseImageParam, AutoImageParam


class BatchedOptimizer:
    def __init__(self, optimizers: List) -> None:
        """Thin abstraction layer to make sure the top level stuff does not change in torch_dreams.Dreamer().render()

        Args:
            optimizers (List): list of torch optimizers
        """
        self.optimizers = optimizers

    def zero_grad(self):
        for opt in self.optimizers:
            opt.zero_grad()

    def step(self):
        for opt in self.optimizers:
            opt.step()


class BatchedImageParam:
    def __init__(
        self,
        image_params: List[BaseImageParam],
    ) -> None:
        self.image_params = image_params

        ## Making sure all of them have the same width and height
        for x in self.image_params:
            assert (
                x.width == self.image_params[0].width
            ), f"Expected widths of all image_params to match, but got: {[z.width for z in self.image_params]}"
            assert (
                x.height == self.image_params[0].height
            ), f"Expected heights of all image_params to match, but got: {[z.height for z in self.image_params]}"

    def forward(self, device: str):
        image_param_batch = torch.cat(
            [x.forward(device=device) for x in self.image_params]
        )
        return image_param_batch

    def clip_grads(self, grad_clip: float = 1.0):
        for p in self.image_params:
            p.clip_grads(grad_clip=grad_clip)


class BatchedAutoImageParam(BatchedImageParam):
    def __init__(
        self,
        batch_size: int = 1,
        height: int = 256,
        width: int = 256,
        standard_deviation: float = 0.01,
        device: str = "cuda:0",
        weight_decay=0.0,
        lr=9e-3,
    ) -> None:
        image_params = [
            AutoImageParam(
                height=height,
                width=width,
                standard_deviation=standard_deviation,
                device=device,
            )
            for x in range(batch_size)
        ]
        for p in image_params:
            p.get_optimizer(lr=lr, weight_decay=weight_decay)

        self.optimizer = BatchedOptimizer(
            optimizers=[x.optimizer for x in image_params]
        )
        super().__init__(image_params)

    def __getitem__(self, idx):
        return self.image_params[idx]

    def __len__(self):
        return len(self.image_params)



from .auto_image_param import BaseImageParam

import cv2 
import torch
import numpy as np

from .constants import Constants
from .error_handlers import PytorchVersionError

from .utils import (
    lucid_colorspace_to_rgb, 
    normalize,
    get_fft_scale_custom_img,
    denormalize,
    rgb_to_lucid_colorspace,
    chw_rgb_to_fft_param,
    fft_to_rgb_custom_img
)

class CustomImageParam(BaseImageParam):
    """FFT parameterization for custom images 

    Works well with:
    * lower learning rates (3e-4) 
    * gradients clipped to (0, 0.1)
    * weight decay (1e-1)

    Args:
        image (str or torch.tensor): 'path/to/image.jpg' or input tensor with shape [1,3, height, width] and values clipped between 0,1.
        device (str): 'cuda' or 'cpu'

    Example: 
    ```
    param = custom_image_param(image = 'image.jpg', device= 'cuda')

    image_param = dreamy_boi.render(
        image_parameter= param,
        layers = [model.Mixed_6c],
        lr = 3e-4,
        grad_clip = 0.1,
        weight_decay= 1e-1
    )

    image_param.save('saved.jpg')
    ```
    """
    def __init__(self, image, device):
        
        super().__init__()
        self.device = device
        if isinstance(image, str):
            image = cv2.cvtColor(cv2.imread(image), cv2.COLOR_BGR2RGB)/255.
            image = torch.tensor(image).permute(-1,0,1).unsqueeze(0)
        self.set_param(image)

    def normalize(self,x, device):
        return normalize(x = x, device= device)

    def postprocess(self, device):
        out = fft_to_rgb_custom_img(height = self.height, width = self.width, image_parameter= self.param, device= device)
        out = lucid_colorspace_to_rgb(t = out, device= device).clamp(0,1)
        return out

    def forward(self, device):
        return self.normalize(self.postprocess(device = device), device= device).clamp(0,1)

    def to_chw_tensor(self, device = 'cpu'):
        t = self.forward(device= device).squeeze(0).clamp(0,1).detach()
        return t
        
    def to_hwc_tensor(self, device = 'cpu'):
        t = self.forward(device= device).squeeze(0).clamp(0,1).permute(1,2,0).detach()
        return t

    def to_nchw_tensor(self, device = 'cpu'):
        """returns a tensor of shape  [1,3, height, width] which represents the image 

        Args:
            device (str, optional): 'cuda' or 'cpu'. Defaults to 'cpu'.

        Returns:
            torch.tensor: NCHW tensor 
        """
        t = self.forward(device= device).clamp(0,1).detach()
        return t

    def set_param(self, tensor):
        """sets an NCHW tensor as the parameter in the frequency domain, 
        useful for transforming custom images between iterations.

        Use in combination with `self.to_nchw_tensor()` like:

        ```
        a = self.to_nchw_tensor()
        # do something with a
        t = transforms.Compose([
            transforms.RandomRotation(45)
        ])
        a = t(a)
        #set as parameter again
        self.set_param(a)
        ```

        WARNING: tensor should have values clipped between 0 and 1. 

        Args:
            tensor (torch.tensor): input tensor with shape [1,3, height, width] and values clipped between 0,1.
        """
        assert len(tensor.shape) == 4

        self.height, self.width = tensor.shape[-2], tensor.shape[-1]
        self.param = chw_rgb_to_fft_param(tensor.squeeze(0), device = self.device)  / get_fft_scale_custom_img(h = self.height, w = self.width, device= self.device)
        self.param.requires_grad_()
        self.optimizer = None



from .dreamer import Dreamer
from .utils import *
from .model_bunch import *
from .tests import *
from .auto_image_param import AutoImageParam
from .custom_image_param import CustomImageParam

__version__ = "3.0.0"

__all__ = [
    "dreamer",
    "utils",
    "model_bunch",
    "auto_image_param",
    "custom_image_param",
    "masked_image_param",
    "image_transforms",
    "transforms"
]



import torch.nn as nn
import torch

from .image_transforms import resize_4d_tensor_by_size


class CaricatureLoss(nn.Module):
    def __init__(self, power=1.0):
        super().__init__()
        self.power = power

    def cosine_dissimilarity(self, x, y, eps=1e-6):
        """
        tried my best to replicate:
            https://github.com/tensorflow/lucid/blob/6dcc927e4ff4e7ef4d9c54d27b0352849dadd1bb/lucid/recipes/caricature.py#L21

        if I missed something out, please get in touch with me on Distill slack: @Mayukh

        or email me:
            mayukhmainak2000@gmail.com
        or find me on github:
            github.com/mayukhdeb
        """

        if x.shape != y.shape:
            """
            if their shapes are not equal (likely due to using static caricatures), then resize the target accordingly
            """
            y = resize_4d_tensor_by_size(
                y.unsqueeze(0), height=x.shape[-2], width=x.shape[-1]
            ).squeeze(0)

        y = y.detach()

        numerator = (x * y.detach()).sum()
        denominator = torch.sqrt((y**2).sum()) + eps
        cossim = numerator / denominator
        cossim = torch.maximum(torch.tensor(0.1).to(cossim.device), cossim)
        loss = -(cossim * numerator**self.power)
        return loss

    def loss(self, x, y):
        loss = self.cosine_dissimilarity(x, y)
        return loss

    def forward(self, layer_outputs, ideal_layer_outputs):
        assert len(layer_outputs) == len(ideal_layer_outputs)
        loss = 0.0
        for i in range(len(layer_outputs)):

            l = self.loss(layer_outputs[i], ideal_layer_outputs[i])
            loss += l

        # print(loss)
        return loss
